--- Instruction_Register (IR) & Decoder for DHBW MCT controller (VHDL)
--- also includes Program Counter (PC)

use work.MCT_Pack_uP.all;
library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

entity IR_DEC is
	port ( Clk: in std_logic; RS : in std_logic; DIN : in P_Type;
	  PC_EN, Pipe_EN, PC_LOADA, PC_LOADR, PC_LOADS,	
	  SP_INC, SP_DEC : in std_logic;
	  A: out DA_Type; -- addresses of operands
	  PSR_Mask : out std_logic_vector(4 downto 0); 
	  -- bit mask for branch operations
	  PC : out PA_Type; OPCODE : out OPTYPE; 
	  K8, StackO : out D_Type; StackI : in D_Type); 
end IR_DEC;

architecture RTL of IR_DEC is
  signal PCINT, SPINT : signed(8 downto 0); 
  -- internal program counter, stack pointer
  signal Offset: unsigned(10 downto 0); 
  signal SP : PA_Type;
  
  begin
  decoder : process (Clk, RS) begin
    if (RS = '1') then OPCODE <= NOP; PSR_Mask <= "00000";
    elsif (Pipe_EN = '1') then
        if rising_edge(Clk) then
        case DIN(15 downto 11) is
          when "00000" => OPCODE <= NOP;
          when "00001" => OPCODE <= LSL;  
          when "00010" => OPCODE <= LSR; 
          when "00011" => OPCODE <= LDI;  K8 <= DIN(10 downto 3);
          when "00100" => OPCODE <= LDA;  A  <= DIN(10 downto 1);          
          when "00101" => OPCODE <= STR;  A  <= DIN(10 downto 1);  
          when "00110" => OPCODE <= ADD;  A  <= DIN(10 downto 1);      
          when "00111" => OPCODE <= SUB;  A  <= DIN(10 downto 1);       
          when "01000" => OPCODE <= ANDA; A  <= DIN(10 downto 1);         
          when "01001" => OPCODE <= EOR;  A  <= DIN(10 downto 1); 
          when "01010" => OPCODE <= ORA;  A  <= DIN(10 downto 1);      
          when "01011" => OPCODE <= JMP;  Offset<=unsigned(DIN(10 downto 0));
          when "01100" => OPCODE <= BRBC; PSR_Mask <= DIN(10 downto 6);
                                          Offset<=unsigned(DIN(10 downto 0));      
          when "01101" => OPCODE <= BRBS; PSR_Mask <= (DIN(10 downto 6));
                                          Offset<=unsigned(DIN(10 downto 0)); 
          when "01110" => OPCODE <= PUSH; A<= "00" & SP; 
          when "01111" => OPCODE <= POP;  A<= '0' & std_logic_vector(SPINT -1); 
          when "10000" => OPCODE <= CALL; Offset<=unsigned(DIN(10 downto 0));
                                          A<= "00" & SP;
                                    StackO <= std_logic_vector(PCINT(7 downto 0)+1);
          when "10001" => OPCODE <= RET;  A<= '0' & std_logic_vector(SPINT -1);          
          when others  => null;
        end case;
      end if;
      end if;
  end process decoder; 
    
  set_PC_and_SP : process (Clk, RS) 
    begin
    if (RS = '1') then PCINT <= (others => '0'); 
                       SPINT <= '0' & x"20"; -- start stack from address 32 
    elsif rising_edge(Clk) then
      if (PC_EN = '1') then 
        if (Pipe_EN = '1') then 
            PCINT <= PCINT + 1;
            -- opcode = PUSH
            if (SP_INC = '1') then SPINT <= SPINT + 1;
            end if;
        -- Pipe_EN = '0'
        elsif (PC_LOADS = '1') then 
            -- opcode = RET (in 2nd cycle when pipe still is disabled)
            PCINT <= signed('0' & StackI(7 downto 0));
        elsif (PC_LOADA = '1') then 
            PCINT <= signed('0' & Offset(7 downto 0));
            -- opcode = CALL
            if (SP_INC = '1') then SPINT <= SPINT + 1; end if;
        elsif (PC_LOADR = '1') then 
            if (Offset(5) = '1') then
              PCINT <= PCINT - signed("0000" & Offset(4 downto 0));
            elsif (Offset(5) = '0') then
              PCINT <= PCINT + signed("0000" & Offset(4 downto 0));
            end if;
        -- opcode = POP (PC_LOADS = '0') & RET (PC_LOADS = '0' in 1st cycle)
        elsif (PC_LOADS = '0') then 
            if (SP_DEC = '1') then SPINT <= SPINT - 1; end if;
        end if;
      end if;
    end if;
  end process set_PC_and_SP;
  
  PC <= std_logic_vector(PCINT(7 downto 0));
  SP <= std_logic_vector(SPINT(7 downto 0));

end RTL;
